//
// CholeskyInverse.h
//
// This file contains the CholeskyInverse generic function.
//
// These classes are part of the Bioinformatics Template Library (BTL).
//
// Copyright (C) 1997,1998 Birkbeck College, Malet Street, London WC1E 7HX, U.K.
// (classlib@mail.cryst.bbk.ac.uk)
// 
// This library is free software; you can
// redistribute it and/or modify it under the terms of
// the GNU Library General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.  This library is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU Library General Public License for more details.
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
///////////////////////////////////////////////////////////////////////////

#if !defined (BTL_CHOLESKYINVERSE_H)
#define BTL_CHOLESKYINVERSE_H

#include "btl/BTL.h"
#include "btl/NumericLimits.h"
#include <math.h>


/**#: [Description ="
    Inverts a symmetric positive definite matrix. Employs Cholesky
    decomposition in step (1). (1)M=LL(tr)   (2)compute L(-1)
    (3)M(-1)=L(-1)(tr)L(-1)
    <P>
    Input should be a symmetric positive definite matrix. Only the lower
    triangular matrix containing the lower triangle of M and overwritten with
    L then with L(-1) and finally with the lower triangle of M(-1)"]
    [Summary = "a generic function to invert a positive definite matrix"] 
    [Authors = "D.S.Moss, W.R.Pitt"]
    [Files = "<A HREF=./btl/CholeskyInverse.h>CholeskyInverse.h</A>"]
    [Dependencies="<A HREF=#BTL_NumericLimits>BTL_NumericLimits</A>"]
    [Prerequisites="The matrix class used as the template parameter must
    contain the following features:<P>
    <P>
    1. value_type - the type of elements in a matrix<P>
    2. size_type  - the subscript type<P>
    3. size_type num_rows(); - function to return the number of rows<P>
    4. size_type num_cols(); - function to return the number of columns<P>
    5. value_type operator()(size_type,size_type); - return individual
       elements of the matrix by the index of the row then the column.
       Indices  should start at 1.<P>"]
*/

template <class Matrix>
class BTL_CholeskyInverse
{
public:
    
    /**#: [Description="Invert the given matrix."] */
    static void 
    Invert(Matrix& m);

};

template<class Matrix>
void BTL_CholeskyInverse<Matrix>::Invert(Matrix& mat)
{
    typedef BTL_TYPENAME Matrix::value_type value_type;
    typedef BTL_TYPENAME Matrix::size_type size_type;
    size_type nrows = mat.num_rows(),
              ncols = mat.num_cols(); 

    value_type a;
    size_type i,j,k;

    if (nrows != ncols)
    	FATAL_ERROR("This Matrix must be square");

    // Compute L where M=LL(tr) using Cholesky decomposition.
    for (i=1; i<=nrows; i++)
    	for (j=1; j<=i; j++)
    	{
	    a=mat(i,j);
	    for (k=1; k<j; k++)
	    	a-=mat(i,k)*mat(j,k);
	    if (i==j)
	    	if (a > BTL_NumericLimits<value_type>::min() * 100)
	    	    mat(i,i)=sqrt(a);
	    	else
	    	{
	    	    WARNING("A principal minor is singular\n");
	    	    mat(i,i)=0.0;
	    	}
	    else
	    	mat(i,j)=a/mat(j,j);
    	}

    // compute L(-1)
    for (j=1; j<=nrows; j++)
    {
    	mat(j,j)=1.0/mat(j,j);
    	for (i=j+1; i<=nrows; i++)
    	{
    	    a=0.0;
    	    for (k=j; k<i; k++)
    	    	a-=mat(i,k)*mat(k,j);
    	    mat(i,j)=a/mat(i,i);
    	}
    }

    // compute M(-1) = L(-1)(tr)L(-1)
    for (j=1; j<=nrows; j++)
    	for (i=j; i<=nrows; i++)
    	{
    	    a=0.0;
    	    for (k=i; k<=nrows; k++)
    	    	a+=mat(k,i)*mat(k,j);
    	    mat(i,j)=mat(j,i)=a;
    	}
    	
} // CholeskyInverse

#endif
