//
// Vector.h
//
// This file contains declarations for the BTL_Vector class.
// This code is part of the Bioinformatics Template Library (BTL).
//
// Copyright (C) 1997,1998 Birkbeck College, Malet Street, London WC1E 7HX, U.K.
// (classlib@mail.cryst.bbk.ac.uk)
//
// This library is free software; you can
// redistribute it and/or modify it under the terms of
// the GNU Library General Public License as published by the Free
// Software Foundation; either version 2 of the License, or (at your
// option) any later version.  This library is distributed in the hope
// that it will be useful, but WITHOUT ANY WARRANTY; without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE.  See the GNU Library General Public License for more details.
// You should have received a copy of the GNU Library General Public
// License along with this library; if not, write to the Free Software
// Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
///////////////////////////////////////////////////////////////////////////



#if !defined(BTL_VECTOR_H)
#define BTL_VECTOR_H


#include "btl/BTL.h"
#include <math.h>
#include <iostream.h>

#if !defined(SGI_CC)
#include <vector>
#else
#include <vector.h>
#endif

#if !defined(SGI_CC) && !defined(GCC_272)
using namespace std;
#endif


/**#: [Description =" his class represents a mathematical vector of any 
    dimension.<P>

                             V = xi + yj + zk + ...<P>

    However, its primary use will probably in representing the position of atoms
    in three dimensions. The default contructor contructs a 3D BTL_Vector.<P>

    There are several types associated with this class:<P>

    <CODE>value_type</CODE> : same as the type BTL_REAL which is defined in
                              BTL.h<P>
    <CODE>iterator</CODE> : this is a STL type iterator. It can be used to
    sequentailly access the elements of the BTL_Vector e.g.<P><CODE>

    	    	    BTL_Vector v1;
    	    	     	...
    	    	    for (BTL_Vector::iterator i = v1.begin(); i != v1.end(); i++)
    	    	    	cout << *i;

    <P>const_iterator</CODE> : used to access const BTL_Vector.<P>
    <CODE>size_type</CODE> : this is an unsigned integer type that represents
    the size of any BTL_Vector object."]
    [Summary = "a vector of real numbers of any dimension"]
    [Authors = "W.R.Pitt, D.S.Moss"]
    [Files = "<A HREF=./btl/Vector.h>Vector.h</A>,
              <A HREF=./btl/Vector.cpp>Vector.cpp</A>"]
    [Friends="Friend equivalents to some functions are available and are
    	      documented with these functions. Also available is: friend
    	      ostream& operator<<(ostream &os, const BTL_Vector &m); "]
    [Dependencies="<A HREF=#BTL_NumericLimits>BTL_NumericLimits</A>
                   <A HREF=#BTL_NumericUtilities>BTL_NumericUtilities</A>"]*/

class BTL_Vector
{
public:

    typedef BTL_REAL			value_type;
    typedef value_type  	    	element_type; // used by TNT
    typedef vector<value_type>  	container;
    typedef container::iterator 	iterator;
    typedef container::const_iterator	const_iterator;
    typedef container::size_type	size_type;

private:

    container vec;   // container - will be changed to the more efficient
	    	    	 // valarray<T> when it becomes available.
public:


    friend class BTL_Matrix;

//.............................................................................
// Constructors/destructor

	    	/**#: [Description="Constructs a 3D vector
	                            and initialised each value to 0.0."]*/

    BTL_Vector() { vec.insert(vec.begin(), 3, 0.0); }

    	    /**#: [Description="Constructs a 3D vector with initialization
    		   and initialises the values with the numbers given."]*/

    BTL_Vector(value_type p, value_type q, value_type r)
    	    { vec.push_back(p); vec.push_back(q); vec.push_back(r); }

    	    /**#: [Description="Constructor for an n dimensional vector.
    				All n values initiated to 0.0."]*/

    BTL_Vector(size_type n) { vec.insert(vec.begin(), n, 0.0);}

    	    /**#: [Description="Constructor for an n dimensional vector.
    				All n values initiated to a value v."]*/

    BTL_Vector(size_type n, value_type v) { vec.insert(vec.begin(), n, v);}

    	    /**#: [Description="Construct from a two dimensional array
    		   of BTL_REALs ( given value_type* )"] */

    BTL_Vector(const value_type* const array, const size_type& p);

    	    /**#: [Description="Copy constructor."]*/

    BTL_Vector(const BTL_Vector &v) : vec(v.vec) {}

    	    /**#: [Description="Destructor."]*/

    ~BTL_Vector() {}

//.............................................................................
// Access functions

    	    /**#: [Description="Returns iterator that can be used to
    		   begin traversing through all elements of the BTL_Vector.
		   (There is also a const version of this function.)"]*/
    iterator
    begin() { return vec.begin(); }

    const_iterator
    begin() const { return vec.begin(); }

    	    /**#: [Description="Returns an iterator can be used in
    		   comparison for ending traversal through the BTL_Vector.
		   (There is also a const version of this function.)"]*/
    iterator
    end() { return vec.end(); }

    const_iterator
    end() const { return vec.end(); }

    	    /**#: [Description="Returns the element, i positions
    		   from the beginning of the BTL_Vector, in constant time.
    		   (i >= 0). (There is also a const version of this
		   function.)"]*/
 
    value_type&
    operator[](const size_type& i);

    value_type
    operator[](const size_type& i) const;


    	    /**#: [Description="Returns the number of elements in the
    		   BTL_Vector."]*/

    size_type
    size() const { return vec.size(); }

	    	/**#: [Description="This function is added to give
		       compatibility with TNT."] */
    size_type
    dim() const
    {
        return vec.size();
    }

    	    /**#: [Description="Returns the i-th element in the BTL_Vector
    		   (i >= 1). (There is a const version of this function as well)."]*/

    value_type&
    operator()(const size_type& i);

    value_type
    operator()(const size_type& i) const {return vec[i-1];}


    	    /**#: [Description="Vector stream output."]*/

    friend ostream&
    operator<<(ostream &os, const BTL_Vector &v);

//.............................................................................
// BTL_Vector comparison

	    	/**#: [Description="Vector equality."]*/

    bool
    operator==(const BTL_Vector& v) const;

    	    /**#: [Description="Returns the largest difference between
    		   equivalent elements in this and BTL_Vector v."]*/

    value_type
    MaxDifference(const BTL_Vector& v) const;

//.............................................................................
// BTL_Vector algebra


    	    /**#: [Description="Vector assignment."]*/

    BTL_Vector&
    operator=(const BTL_Vector &v) { vec = v.vec;return *this;}


    	    /**#: [Description="Vector addition."]*/

    BTL_Vector
    operator+(const BTL_Vector &v) const;


    	    /**#: [Description="Vector subtraction."]*/

    BTL_Vector
    operator-(const BTL_Vector &v) const;


    	    /**#: [Description="Vector negation (unary minus operator)."]*/

    BTL_Vector
    operator-() const;


    	    /**#: [Description="Vector increment."]*/

    BTL_Vector&
    operator+=(const BTL_Vector &v);


    	    /**#: [Description="Vector decrement."]*/

    BTL_Vector&
    operator-=(const BTL_Vector &v);


    	    /**#: [Description="Add a scalar to each element in
    		   the BTL_Vector."]*/

    BTL_Vector
    operator+(const value_type &s) const;


    	    /**#: [Description="Subtract a scalar from each element in
    		   the BTL_Vector."]*/
 
    BTL_Vector
    operator-(const value_type &s) const;


    	    /**#: [Description="Add a scalar to each element in
    		   the BTL_Vector."]*/
 
    BTL_Vector&
    operator+=(const value_type &s);


    	    /**#: [Description="Subtract a scalar from each element in
    		   the BTL_Vector."]*/
 
    BTL_Vector&
    operator-=(const value_type &s);


    	    /**#: [Description="Vector scalar/dot product."]*/
 
    value_type
    operator*(const BTL_Vector &v) const;


    	    /**#: [Description="Vector cross product."]*/
 
    BTL_Vector
    operator%(const BTL_Vector &v) const;


    	    /**#: [Description="Vector multiplication by a scalar."]*/
 
    BTL_Vector
    operator*(const value_type& y) const;


    	    /**#: [Description="Vector multiplication by a scalar."]*/
 
    friend BTL_Vector
    operator*(const value_type& y, const BTL_Vector &v) { return (v * y); }


    	    /**#: [Description="Vector multiplication by a scalar and
    		  assignment."]*/
 
    BTL_Vector&
    operator*=(const value_type& y);


    	    /**#: [Description="Vector division by a scalar."]*/
 
    BTL_Vector
    operator/(const value_type& y) const;

    	    /**#: [Description="Vector division by a scalar."]*/
 
    friend BTL_Vector
    operator/(const value_type& y, const BTL_Vector &v) { return (v / y); }


    	    /**#: [Description="Vector division by a scalar and
    		   assignment."]*/
 
    BTL_Vector&
    operator/=(const value_type& y);

    	    /**#: [Description="Returns the root mean square distance
    		   between this and BTL_Vector v."]*/
 
    value_type
    Distance(const BTL_Vector& v) const;


    	    /**#: [Description="Returns zero if the cooresponding BTL_Vector
    		   elements are all identical within the floating point
    		   precision used. If they are not identical then the
    		   function finds the largest difference between
    		   cooresponding elements and outputs the number of
    		   significant digits in this difference. The range this
    		   number can take depends on the floating point precision
    		   used. The maximum is typically equal to 6 for
    		   single precision (FLT_DIG) and 15 for double precision
    		   (DBL_DIG)."]*/
 
    unsigned int
    IsNotEqualTo(const BTL_Vector& v) const;

    	    /**#: [Description="Returns the modulus/magnitude of the
    				vector."]*/
 
    value_type
    Modulus() const;

};

#endif // BTL_Vector.h
