// MatrixTest.cpp - demo/test routines for Matrix member functions.

#include "btl/Matrix.h"
#include "btl/NumericUtilities.h"
#include "btl/LeastSquaresFit.h"
#include <iomanip.h>

//.............................................................................

void CheckCompare(BTL_Matrix& m)
{
    cout<<"\n#####################################\n"
    	 << "### Check Compare()              ####\n"
    	 << "###                              ####\n"
    	 << "### Fitting a matrix to itself   ####\n"
    	 << "#####################################\n";


    // Copy input to matrix a.
    BTL_Matrix a = m;

    // Define types to be used
    typedef BTL_Matrix::iterator iterator;
    typedef BTL_Matrix::const_iterator const_iterator;
    typedef BTL_LeastSquaresFit<iterator,const_iterator> Fitter;


    // (1)
    //
    // Calculate root mean squared deviation after a least squares fit of a to m
    // and output the result
    BTL_Matrix::value_type 
    RMSD = Fitter::Compare(a.begin(),a.end(),m.begin(),m.end());

    cout << "\nRMSD's (should be close to zero)\n\n";
    cout << setprecision(16) << "After fitting to itself : " << RMSD << '\n';

    // (2)
    //
    // Translate input matrix then fit to the original.
    BTL_Vector trans(43.1,8.0,-9.0);
    a = m + trans;

    RMSD = Fitter::Compare(a.begin(),a.end(),m.begin(),m.end());
                                                             
    cout << "After translation, then fitting : " << RMSD << '\n';

    // (3)
    //
    // Rotate and translate input matrix then fit to the original.
    BTL_Vector axis(45.0,56.9,-9999.9);
    BTL_Matrix rot(axis,45.0);
    a = m % rot + trans;

    RMSD = Fitter::Compare(a.begin(),a.end(),m.begin(),m.end());
                                                             
    cout << "After rotation and translation, then fitting : " << RMSD 
         << "\n\n";
         
}
//.............................................................................

void CheckFit(BTL_Matrix& m)
{
    cout<<"\n####################################\n"
    	 << "### Check Fit()                 ####\n"
    	 << "###                             ####\n"
    	 << "### Fitting a matrix to itself  ####\n"
    	 << "####################################\n";

    // Copy input to matrix a.
    BTL_Matrix a = m;

    // Define types to be used
    typedef BTL_Matrix::iterator iterator;
    typedef BTL_Matrix::const_iterator const_iterator;
    typedef BTL_LeastSquaresFit<iterator,const_iterator> Fitter;


    // (1)
    //
    // Translate input matrix then fit to the original.
    BTL_Vector trans(0.0,8.0,-9.0);
    a = m + trans;

    double RMSD = Fitter::Fit(a.begin(),a.end(),m.begin(),m.end());

    cout << setprecision(16) << "\nRMSD's (should be close to zero)\n";
    cout << "\nRMSD after translation then fitting (Kearsley's method) : " 
         << RMSD;

    // Calculate RMSD using normal method
    RMSD = Fitter::RMSD(a.begin(),a.end(),m.begin(),m.end(),3);

    cout << "\nRMSD after translation then fitting (explicit method)   : " 
         << RMSD;

    // (2)
    //
    // Rotate and translate input matrix then fit to the original.
    BTL_Vector axis(45.0,56.9,-99.9);
    BTL_Matrix rot(axis,45.0);
    a = m % rot + trans;
    
    RMSD = Fitter::Fit(a.begin(),a.end(),m.begin(),m.end());
    
    cout << "\nRMSD after rotation, translation then fitting "
         << "(Kearsley's method) : " << RMSD;

    RMSD = Fitter::RMSD(a.begin(),a.end(),m.begin(),m.end(),3);

    cout << "\nRMSD after rotation, translation then fitting "
         << "(explicit method)   : " << RMSD << "\n\n";
}

//.............................................................................

template<class iterator, class const_iterator>
void CheckFit(iterator begin1, iterator end1, 
              const_iterator begin2, const_iterator end2)
{
    cout<<"\n###############################\n"
    	 << "### Check Fit()            ####\n"
    	 << "### Fitting insulin (4ins) ####\n"
    	 << "### chain A to chain C     ####\n"
    	 << "###############################\n";

    // Abbrieviate the name of the fitting routine
    typedef BTL_LeastSquaresFit<iterator,const_iterator> Fitter;

    // Do fit
    double RMSD = Fitter::Fit(begin1,end1,begin2,end2);

    cout << "\nRMSD (Kearsley's method) : "  << RMSD
         << " (Quanta and Profit 1.314)";

    // Check RMSD using normal method of calculation 
    RMSD = Fitter::RMSD(begin1,end1,begin2,end2,3);

    cout << "\nRMSD (explicit method)   : "  << RMSD << "\n\n";
}
//.............................................................................

template<class iterator, class const_iterator>
void CheckCompare(iterator begin1, iterator end1, 
                  const_iterator begin2, const_iterator end2)
{
    cout<<"\n###############################\n"
    	 << "### Check Compare()        ####\n"
    	 << "### Fitting insulin (4ins) ####\n"
    	 << "### chain A to chain C     ####\n"
    	 << "###############################\n";

    // Abbrieviate the name of the fitting routine
    typedef BTL_LeastSquaresFit<iterator,const_iterator> Fitter;

    // Do fit
    double RMSD = Fitter::Compare(begin1,end1,begin2,end2);

    cout << "\nRMSD (Kearsley's method) : "  << RMSD 
         << " (Quanta and Profit 1.314)\n\n";
}

//.............................................................................

template<class iterator, class const_iterator>
void CheckRMSD(iterator begin1, iterator end1, 
               const_iterator begin2, const_iterator end2)
{
    cout<<"\n###############################\n"
    	 << "### Check RMSD()           ####\n"
    	 << "### Using insulin (4ins)   ####\n"
    	 << "### chain A to chain C     ####\n"
    	 << "###############################\n";

    // Abbrieviate the name of the fitting routine
    typedef BTL_LeastSquaresFit<iterator,const_iterator> Fitter;

    // Do calculation
    double RMSD = Fitter::RMSD(begin1,end1,begin2,end2,3);

    cout << "\nRMSD (explicit method) : "  << RMSD 
         << " (Quanta 29.44 for 161 atoms)\n\n";
}

//.............................................................................

int main()
{

    // Set up data in the form of two 1D arrays - with pointers to the 
    // beginning and end of each array
    //
#include "../testCode/p4ins.dat"
    BTL_REAL *beginA = p4ins_A, *endA = beginA + 163*3,
         *beginC = p4ins_C, *endC = beginC + 163*3;

    // (1)
    //
    // Test fitting routines: fitting Matrix to itself after rigid body
    // movements
    //
    BTL_Matrix matA(p4ins_A,163,3);
    BTL_Matrix matC(p4ins_C,163,3);
    CheckCompare(matA);
    CheckFit(matA);

    // (2)
    //
    // Check fitting of two sets of coordinates: in Matrix form
    //
    CheckRMSD(matA.begin(),matA.end(),matC.begin(),matC.end());
    CheckCompare(matA.begin(),matA.end(),matC.begin(),matC.end());
    CheckFit(matA.begin(),matA.end(),matC.begin(),matC.end());

    // (3)
    //
    // Test with STL vectors
    //
    vector<BTL_REAL> vecA(beginA,endA), vecC(beginC,endC);

    cout << "\n********* Repeat last 3 tests using STL vectors "
         << "*********\n\n";

    CheckRMSD(vecA.begin(),vecA.end(),vecC.begin(),vecC.end());
    CheckCompare(vecA.begin(),vecA.end(),vecC.begin(),vecC.end());
    CheckFit(vecA.begin(),vecA.end(),vecC.begin(),vecC.end());

    // (4)
    //
    // Test with 1D arrays instead of Matrix's
    //

    cout << "\n********* Repeat last 3 tests using arrays instead of Matrix's "
         << "*********\n\n";

    CheckRMSD(beginA,endA,beginC,endC);
    CheckCompare(beginA,endA,beginC,endC);
    CheckFit(beginA,endA,beginC,endC);

    return 0;
}
