// BTL_MatrixTest.cpp - demo/test routines for Matrix member functions.


#include "btl/BTL.h"

#if !defined(SGI_CC)
#include <string>
#else
#include <mstring.h>
#endif

#include "btl/Matrix.h"
#include "btl/NumericUtilities.h"
#include "btl/CholeskyInverse.h"
#include <iomanip.h>

#if !defined(SGI_CC) && !defined(GCC_272)
using namespace std;
#endif

//.............................................................................

void TestResults(const string& testName, const unsigned int& error)
{
    cout << ">>>>> ERROR = " 
    	 << error
    	 << " : "
    	 << testName 
    	 << " <<<<<\n" 
    	 << endl;
}

//.............................................................................

BTL_Vector::value_type CreateReal()
{
    // Set different random number seed each time this function is called.
    static unsigned int seed = 0;
    seed++;
    srand(seed);
    return (BTL_Vector::value_type) rand();
}

//.............................................................................

BTL_Vector CreateVector(BTL_Vector::size_type size)
{
    // Set different random number seed each time this function is called.
    static unsigned int seed = 0;
    seed++;
    srand(seed);

    BTL_Vector vec(size);
    for (BTL_Vector::size_type i=0; i<size; i++)
    	vec[i] = (BTL_Vector::value_type) rand();
    return vec;
}

//.............................................................................

BTL_Matrix CreateMatrix(BTL_Matrix::size_type nrows, 
                        BTL_Matrix::size_type ncols)
{
    // Set different random number seed each time this function is called.
    static unsigned int seed = 0;
    seed++;
    srand(seed);

    BTL_Matrix mat(nrows,ncols);
    for (BTL_Matrix::iterator i=mat.begin(); i!=mat.end(); i++)
    	*i = (BTL_Matrix::value_type) rand();
    return mat;
}

//.............................................................................
// Squares a matrix by removing excess columns or rows

BTL_Matrix Square(BTL_Matrix& m)
{
    typedef BTL_Matrix::size_type size_type;
    
    size_type nrows = m.ReadNRows(),
              ncols = m.ReadNCols(),
              size;
        
    if (nrows == ncols)
    	return m;

    if (nrows > ncols) size = ncols; else size = nrows;
    
    BTL_Matrix squareMat(size,size);
    
    for (size_type row=1; row<=size; row++)   
    	for (size_type col=1; col<=size; col++)   
    	    squareMat(row,col) = m(row,col);
    	    
    return squareMat;
}
//.............................................................................
// Symmetrifys a square matrix by copying the upper triangle into the lower
// triangle

BTL_Matrix Symmetrify(BTL_Matrix& m)
{
    typedef BTL_Matrix::size_type size_type; 

    BTL_Matrix msqr = Square(m);

    size_type size = msqr.ReadNRows();

    for (size_type row=1; row<=size; row++)
    	for (size_type col=1; col<=size; col++)
     	    msqr(col,row) = msqr(row,col);
     	    
    return msqr;
}

//.............................................................................
// Trims a matrix to a required size.

BTL_Matrix Trim(BTL_Matrix& m, unsigned long rowMax, unsigned long colMax)
{
    typedef BTL_Matrix::size_type size_type;
    
    size_type oldNrows = m.ReadNRows(),
              oldNcols = m.ReadNCols();
              
    size_type nrows, ncols;
    if (rowMax < oldNrows) nrows = rowMax; else nrows = oldNrows; 
    if (colMax < oldNcols) ncols = colMax; else ncols = oldNcols;
    
    BTL_Matrix trimmedMat(nrows,ncols);
    
    for (size_type row=1; row<=nrows; row++)   
    	for (size_type col=1; col<=ncols; col++)   
    	    trimmedMat(row,col) = m(row,col);
    	    
    return trimmedMat;
}

//.............................................................................
// Finds whether a Matrix is positive definite
     
bool PositiveDefinite(BTL_Matrix& m)
{
    // For a matrix to be positive definite, all the determinants of
    // matrices of size 1, 2x2, 3x3, ... nxn where n is the dimension of the 
    // original must be greater than zero.

    typedef BTL_Matrix::value_type value_type;
    typedef BTL_Matrix::size_type size_type;
    
    size_type size = m.ReadNRows();
    for (size_type i=1; i<=size; i++)
    {
    	BTL_Matrix temp(i,i);
    	for (size_type row=1; row<=i; row++)
    	    for (size_type col=1; col<=i; col++)
    	    	temp(row,col) = m(row,col);
    	value_type determinant = temp.Determinant();
    	if (determinant < BTL_NumericLimits<value_type>::epsilon())
    	    return false;
    }
    return true;    	
}

//.............................................................................

void CheckConstrct1()
{
    cout<<"\n############################################\n"
    	 << "### Checking matrix constructors1:      ####\n"
    	 << "###                                     ####\n"
    	 << "### constructing 3x3 matrices of zeros. ####\n"
    	 << "############################################\n\n";

    BTL_Matrix a;

    BTL_Matrix b(3,3);

    unsigned int error = a.IsNotEqualTo(b);
    TestResults("Constructing 3x3 zero Matrix's.",error);

    BTL_Matrix c(1.1, 1.2, 1.3,
             2.1, 2.2, 2.3,
             3.1, 3.2, 3.3);

    BTL_Matrix::value_type darr[9] = {1.1, 1.2, 1.3,
 				  2.1, 2.2, 2.3,
 				  3.1, 3.2, 3.3};
    BTL_Matrix d(darr,3,3);

    error = c.IsNotEqualTo(d);
    TestResults("Constructing 3x3 Matrix's with non-zero elements.",error);

    BTL_Matrix::value_type earr[3][3] = {{1.1, 1.2, 1.3},
 				     {2.1, 2.2, 2.3},
 				     {3.1, 3.2, 3.3}};

    BTL_Matrix e(&earr[0][0],3,3);

    error = c.IsNotEqualTo(e);
    TestResults("Constructing a 3x3 Matrix from a 2D array.",error);

    BTL_Matrix f(3,3);

    for (BTL_Matrix::size_type i=0; i!=3; i++)
    	f.WriteRow(&earr[i][0],i+1);

    error = e.IsNotEqualTo(f);
    TestResults("Writing to a Matrix using WriteRow.",error);

}
//.............................................................................

void CheckConstrct2(BTL_Matrix& m)
{
    cout<<"\n############################################\n"
    	 << "### Checking matrix constructors2:      ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";

    typedef BTL_Matrix::size_type size_type;
    typedef BTL_Matrix::value_type value_type;
    typedef BTL_Matrix::iterator iterator;

    size_type nrows = m.ReadNRows(),
                      ncols = m.ReadNCols(),
                      size  = ncols * nrows;

    // Copy contents to a 1D array and then, from there, to another matrix
    value_type *aarr = new value_type[size];
    iterator miter=m.begin();
    for (size_type i=0; i<size; i++, miter++)
    	aarr[i] = *miter;
    BTL_Matrix a(aarr,nrows,ncols);

    unsigned int error = a.IsNotEqualTo(m);
    TestResults("Constructing a BTL_Matrix from a 1D array.",error);

    delete [] aarr;

    BTL_Matrix b = a;

    error = b.IsNotEqualTo(m);
    TestResults("Using the copy constructor.",error);
}
//.............................................................................

void CheckRotateMat(BTL_Vector& point, BTL_Vector& axis, float numOfRots)
{
    cout<<"\n############################################\n"
    	 << "### Checking construction of a rotation ####\n"
    	 << "### matrix                              ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";

    typedef BTL_Matrix::size_type size_type;
    typedef BTL_Matrix::value_type value_type;

    value_type angle = 360.0/numOfRots;

    // Rotate in 360 degrees in +ve direction
    BTL_Matrix rot1(axis,angle);
    BTL_Vector movable = point;
    // extra curly brackets are for the benefit of SGI CC version 7.1
    {for (size_type i=1; i<=numOfRots; i++)
    	movable = rot1 *  movable;}

    // Rotate in 360 degrees in -ve direction
    BTL_Matrix rot2(axis,-angle);
    rot2 = rot2.Transpose();
    for (size_type i=1; i<=numOfRots; i++)
    	movable = movable * rot2;

    unsigned int error = point.IsNotEqualTo(movable);
    TestResults("Constructing a rotation Matrix and applying it.",error);
}
//.............................................................................


void CheckTranspose(BTL_Matrix& m)
{
    cout<<"\n#########################################\n"
    	 << "### Checking Transpose()             ####\n"
    	 << "###                                  ####\n"
    	 << "### m == m.Transpose().Transpose() ? ####\n"
    	 << "#########################################\n\n";

    // Calculate transpose

    BTL_Matrix mT = m.Transpose();

    // Check that transpose of the transpose equals the original.

    unsigned int error = m.IsNotEqualTo(mT.Transpose());
    TestResults("Matrix transpose.",error);
}
//.............................................................................

void CheckSquareInverse(BTL_Matrix& m)
{
    cout<<"\n################################################\n"
    	 << "### Checking SquareInverse()                ####\n"
    	 << "###                                         ####\n"
    	 << "### m == m.SquareInverse().SquareInverse()? ####\n"
    	 << "################################################\n\n";

    // Calculate inverse

    BTL_Matrix msymm = Square(m);
    BTL_Matrix mI = msymm;
    mI.SquareInverse();
    mI.SquareInverse();

    unsigned int error = msymm.IsNotEqualTo(mI);
    TestResults("Inverse of small square matrix.",error);
}

//.............................................................................

void CheckInverse(BTL_Matrix& m)
{
    cout<<"\n################################################\n"
    	 << "### Checking Inverse()                      ####\n"
    	 << "###                                         ####\n"
    	 << "### A * A.Inverse() * A == A ?              ####\n"
    	 << "################################################\n\n";

    // Calculate inverse

    BTL_Matrix mInv = m;
    mInv.Inverse();
    BTL_Matrix lhs = m * mInv * m;
    
    unsigned int error = lhs.IsNotEqualTo(m);
    TestResults("A * A.Inverse() * A == A?",error);
}
//.............................................................................

void CheckUnity(BTL_Matrix& m)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix multiplication   ####\n"
    	 << "###  and unit matrix constructor     ####\n"
    	 << "###                                  ####\n"
    	 << "### unity = m * m.SquareInverse() ?  ####\n"
    	 << "#########################################\n\n";

    // Calculate inverse

    BTL_Matrix msymm = Square(m);
    BTL_Matrix mI = msymm.SquareInverse();

    // Construct a unit matrix;

    BTL_Matrix unity(msymm.ReadNRows());

    // Check that the original * its inverse equals the unit matrix.
    
    BTL_Matrix rhs = msymm * mI;
    
    BTL_Matrix::value_type maxDiff = unity.MaxDifference(rhs);
    
    cout << "\nThe maximum difference between the two sides of the above\n"
         << "equation (should be close to zero) : " 
         << setprecision(30) 
         << maxDiff
         << "\n\n";   
}
//.............................................................................

void CheckOperators1(const BTL_Matrix& m)
{
    cout<<"\n#########################################\n"
    	 << "### Checking overloaded operators    ####\n"
    	 << "###    the const () and []  and =    ####\n"
    	 << "#########################################\n\n";

    BTL_Matrix a;
    a = m;

    unsigned int error = a.IsNotEqualTo(m);
    TestResults("The = operator.",error);
 
    // check ()
    typedef BTL_Matrix::size_type size_type;

    size_type nrows = m.ReadNRows(),
    	      ncols = m.ReadNCols();
    
    unsigned int maxError=0;
    typedef BTL_Matrix::value_type value_type;
    for (size_type row=1; row<=nrows; row++)
    	for (size_type col=1; col<=ncols; col++)
    	{
    	    error = 
    	       BTL_NumericUtilities<value_type>::AreNotEqual(a(row,col),m(row,col));
    	    if (error > maxError) maxError=error;
        }
    TestResults("The const () operator.",maxError);
 
    // check []
    maxError = 0;
    for (size_type i=0; i<nrows; i++)
    	for (size_type j=0; i<ncols; i++)
    	{
    	    error = BTL_NumericUtilities<value_type>::AreNotEqual(a[i][j],m[i][j]);
    	    if (error > maxError) maxError=error;
    	}

    TestResults("The const [] operator.",error);
}
 
//.............................................................................

void CheckOperators2(BTL_Matrix& m)
{
    cout<<"\n#########################################\n"
    	 << "### Checking overloaded operators    ####\n"
    	 << "###    the  + += - -= * *= / /=      ####\n"
    	 << "#########################################\n\n";
    	 
    // 	Vector operator*(const BTL_Vector &v) const; and
    // 	friend BTL_Vector operator*(const BTL_Vector &v, const BTL_Matrix& m);
    //  checked in CheckRotateMat.

    //  BTL_Matrix operator*(const BTL_Matrix& m) const;
    //  checked in CheckTransMult.

    //  check multiplication and division by scalar.

    BTL_Matrix b = m;
    typedef BTL_Matrix::value_type value_type;
    value_type scale = -99999.9089;
    b *= scale;
    b = b * scale;
    b = scale * b;
    b /= scale;
    b = b / scale;
    b /= scale;

    unsigned int error = b.IsNotEqualTo(m);
    TestResults("The *= * /= and / operators : between BTL_Matrix's and "
                "scalars.", error);
 
    // check addition and subtraction by a scalar.

    BTL_Matrix c = m;
    value_type shift = -99999.9089;
    c += shift;
    c = c + shift;
    c -= shift;
    c = c - shift;

    error = c.IsNotEqualTo(m);
    TestResults("The += + -= and - operators : between BTL_Matrix's and scalars.",
                error);

    // check row addition and subtraction by a vector.

    BTL_Matrix d = m;
    BTL_Vector randVec = CreateVector(m.ReadNCols());
    	
    d += randVec;
    d = d + randVec;
    d -= randVec;
    d = d - randVec;

    error = d.IsNotEqualTo(m);
    TestResults("The += + -= and - operators : between Matrix's and BTL_Vector's.",
                error);

    // check matrix addition and subtraction.

    BTL_Matrix randMat = CreateMatrix(m.ReadNRows(),m.ReadNCols());

    BTL_Matrix e = m;
    e += randMat;
    e = e + randMat;
    e -= randMat;
    e = e - randMat;

    error = e.IsNotEqualTo(m);
    TestResults("The += + -= and - operators : between Matrix's.", error);
}

//.............................................................................

void CheckTransMult(BTL_Matrix& m)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix multiplication   ####\n"
    	 << "###                                  ####\n"
    	 << "#########################################\n\n";

    // Calculate inverse

    // Construct a second matrix from the first one.
    BTL_Matrix m2 = m.Transpose();
    typedef BTL_Matrix::size_type size_type;
    size_type nrows = m2.ReadNRows(), ncols = m2.ReadNCols();
    for (size_type i=1; i<=nrows; i++)
         for (size_type j=1; j<=ncols; j++)
    	         m2(i,j) = (i + 0.75) * (j - 10.0);

    BTL_Matrix lhs = (m * m2).Transpose();

    BTL_Matrix rhs1 = m2.Transpose() * m.Transpose();
    unsigned int error = lhs.IsNotEqualTo(rhs1);
    TestResults("(m1*m2).Transpose() == m2.Transpose() * m1.Transpose ?",error);

    BTL_Matrix rhs2 = m2.Transpose() % m;
    error = lhs.IsNotEqualTo(rhs2);
    TestResults("(m1*m2).Transpose() == m2.Transpose() % m1 ?", error);

    BTL_Matrix rhs3 = m2 & m.Transpose();
    error = lhs.IsNotEqualTo(rhs3);
    TestResults("(m1*m2).Transpose() == m2 & m1.Transpose() ?", error);
}
//.............................................................................

void CheckProperty1(BTL_Matrix& m1, BTL_Matrix& m2)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix properties 1     ####\n"
    	 << "###                                  ####\n"
    	 << "#########################################\n\n";
    	 
    typedef BTL_Matrix::size_type size_type;

    size_type m1Nrows = m1.ReadNRows(),
    	      m2Nrows = m2.ReadNRows(),
    	      m1Ncols = m1.ReadNCols(),
    	      m2Ncols = m2.ReadNCols();

    size_type nrows,ncols;
    BTL_Matrix A,B;

    // Restrictions: A.size == B.size.
    nrows = ( m1Nrows < m2Nrows ? m1Nrows : m2Nrows );
    ncols = ( m1Ncols < m2Ncols ? m1Ncols : m2Ncols );
    A = Trim(m1,nrows,ncols);
    B = Trim(m2,nrows,ncols);
    
    unsigned int error = (A + B).IsNotEqualTo(B + A);

    TestResults("A + B == B + A ?", error);
}
//.............................................................................

void CheckProperty2(BTL_Matrix& m1, BTL_Matrix& m2, BTL_Matrix& m3)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix properties 2     ####\n"
    	 << "###                                  ####\n"
    	 << "#########################################\n\n";

    typedef BTL_Matrix::size_type size_type;

    size_type m1Nrows = m1.ReadNRows(),
    	      m2Nrows = m2.ReadNRows(),
    	      m3Nrows = m3.ReadNRows(),
    	      m1Ncols = m1.ReadNCols(),
    	      m2Ncols = m2.ReadNCols(),
    	      m3Ncols = m3.ReadNCols();

    size_type nrows,ncols;
    BTL_Matrix A,B,C;

    // Restrictions: A.size == B.size == C.size.
    nrows = m1Nrows;
    ncols = m1Ncols;
    if (m2Nrows < nrows) nrows = m2Nrows;
    if (m3Nrows < nrows) nrows = m3Nrows;
    if (m2Ncols < ncols) ncols = m2Ncols;
    if (m3Ncols < ncols) ncols = m3Ncols;
    A = Trim(m1,nrows,ncols);
    B = Trim(m2,nrows,ncols);
    C = Trim(m3,nrows,ncols);

    unsigned int error = (A + B + C).IsNotEqualTo(A + ( B + C ));
    TestResults("(A + B) + C == A + (B + C) ?", error);
}
//.............................................................................

void CheckProperty3(BTL_Matrix& m1, BTL_Matrix& m2, BTL_Matrix& m3)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix properties 3     ####\n"
    	 << "###                                  ####\n"
    	 << "#########################################\n\n";

    typedef BTL_Matrix::size_type size_type;

    size_type m2Nrows = m2.ReadNRows(),
    	      m3Nrows = m3.ReadNRows(),
    	      m1Ncols = m1.ReadNCols(),
    	      m2Ncols = m2.ReadNCols(),
    	      m3Ncols = m3.ReadNCols();

    size_type nrows,ncols;
    BTL_Matrix A,B,C;

    // Restrictions: B.size == C.size &&
    //               A.ncols == B.nrows;
    nrows = m1Ncols;
    if (m2Nrows < nrows) nrows = m2Nrows;
    if (m3Nrows < nrows) nrows = m3Nrows;
    ncols = m2Ncols;
    if (m3Ncols < ncols) ncols = m3Ncols;

    A = Trim(m1,99999999,nrows);
    B = Trim(m2,nrows,ncols);
    C = Trim(m3,nrows,ncols);
    BTL_Matrix lhs = A * (B + C);
    BTL_Matrix rhs = A * B + A * C;

    unsigned int error = lhs.IsNotEqualTo(rhs);
    TestResults("A * (B + C) == A * B + A * C ?", error);
}
//.............................................................................

void CheckProperty4(BTL_Matrix& m1, BTL_Matrix& m2, BTL_Matrix& m3)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix properties 4     ####\n"
    	 << "###                                  ####\n"
   	 << "#########################################\n\n";

    typedef BTL_Matrix::size_type size_type;

    size_type m2Nrows = m2.ReadNRows(),
    	      m3Nrows = m3.ReadNRows(),
    	      m1Ncols = m1.ReadNCols(),
    	      m2Ncols = m2.ReadNCols();

    size_type nrows,ncols;
    BTL_Matrix A,B,C;


    // Restrictions: A.ncols == B.nrows && B.ncols == C.nrows
    ncols = m1Ncols;
    if (m2Nrows < ncols) ncols = m2Nrows;
    nrows = m3Nrows;
    if (m2Ncols < nrows) nrows = m2Ncols;
    A = Trim(m1,99999999,ncols);
    B = Trim(m2,ncols,nrows);
    C = Trim(m3,nrows,99999999);

    BTL_Matrix lhs = A * (B * C);
    BTL_Matrix rhs = A * B * C;

    unsigned int error = lhs.IsNotEqualTo(rhs);
    TestResults("A * (B * C) == A * B * C ?", error);
}

//.............................................................................

void CheckProperty5(BTL_Matrix& m1, BTL_Matrix& m2)
{
    cout<<"\n#########################################\n"
    	 << "### Checking matrix properties 5     ####\n"
    	 << "###                                  ####\n"
    	 << "#########################################\n\n";

    typedef BTL_Matrix::size_type size_type;

    size_type m1Nrows = m1.ReadNRows(),
    	      m2Nrows = m2.ReadNRows(),
    	      m1Ncols = m1.ReadNCols(),
    	      m2Ncols = m2.ReadNCols();

    size_type nrows,ncols;
    BTL_Matrix A,B;

    // Restrictions: A.size == B.size.
    nrows = ( m1Nrows < m2Nrows ? m1Nrows : m2Nrows );
    ncols = ( m1Ncols < m2Ncols ? m1Ncols : m2Ncols );
    A = Trim(m1,nrows,ncols);
    B = Trim(m2,nrows,ncols);

    BTL_Matrix lhs = (A + B).Transpose(),
    	   rhs = A.Transpose() + B.Transpose();
    	   
    unsigned int error = lhs.IsNotEqualTo(rhs);
    TestResults("(A + B).Transpose() == A.Transpose() + B.Transpose() ?",error);
}
//.............................................................................

void CheckEigens(BTL_Matrix& m)
{
    cout<<"\n############################################\n"
    	 << "### Check Eigens()                      ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";

    BTL_Matrix msymm = Symmetrify(m);
    typedef BTL_Matrix::size_type size_type;
    size_type size = msymm.ReadNRows();
    BTL_Vector eigenValues(size);
    BTL_Matrix eigenVectors(size,size);
    msymm.Eigens(eigenValues,eigenVectors);
//    BTL_Matrix transMult = eigenVectors % eigenVectors;
//    unsigned int error = transMult.IsNotEqualTo(unity), maxError = error;
    BTL_Matrix unity(size);


    // ME = eE; where M is the original matrix (msymm), E is an eigen vector
    //          and e is its corresponding eigen value.

    // Multiple each eigen vector by the original matrix;
    // V = ME; where V is one column in temp.
    BTL_Matrix temp = msymm * eigenVectors;

    // Divide resulting vectors by the corresponding part of an eigen vector.
    // e(i) = V(i)/E(i); where V is one column of temp and E is its
    //                   corresponding eigen vector.

    // extra curly brackets are for the benefit of SGI CC version 7.1
    {for (size_type col=1; col<=size; col++)
    	for (size_type row=1; row<=size; row++)
    	    temp(row,col) /= eigenVectors(row,col);}

    unsigned int error=0, maxError=0;
    typedef BTL_Matrix::value_type value_type;
    for (size_type col=1; col<=size; col++)
    	for (size_type row=1; row<=size; row++)
    	{
    	    value_type lhs = temp(row,col), rhs = eigenValues(col);
    	    error = BTL_NumericUtilities<value_type>::AreNotEqual(lhs,rhs);
    	    if (error > maxError) maxError=error;
        }

    TestResults("Eigen vectors and Eigen values ?", error);
}

void CheckCholesky(BTL_Matrix& m)
{
    cout<<"\n############################################\n"
    	 << "### Check CholeskyInverse()             ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";
    	 
    BTL_Matrix msqr = Symmetrify(m);
    bool posdef = PositiveDefinite(msqr);
    if (!posdef)
    	cout << "!Problem in CheckCholesky(BTL_Matrix& m)\n"
    	     << "Input matrix not compatible with this test.\n\n";

    BTL_Matrix minv = msqr;
    BTL_CholeskyInverse<BTL_Matrix>::Invert(minv);
    BTL_CholeskyInverse<BTL_Matrix>::Invert(minv);
    
    unsigned int error = msqr.IsNotEqualTo(minv);

    TestResults("BTL_CholeskyInverse::Invert() ?",error);
}
//.............................................................................
    
void CheckRow(BTL_Matrix& m)
{
    cout<<"\n############################################\n"
    	 << "### Check ReadRow and WriteRow          ####\n"
    	 << "###                                     ####\n"
    	 << "############################################\n\n";
    	 
    typedef BTL_Matrix::size_type size_type;

    size_type ncols = m.ReadNCols();
    size_type nrows = m.ReadNRows();
    
    vector<BTL_Vector> rowvecs;
    for (size_type i=1; i<=nrows; i++)
    {
    	BTL_Vector rowi = m.ReadRow(i);
    	rowvecs.push_back(rowi);
    }
    
    typedef vector<BTL_Vector>::iterator iterator;

    BTL_Matrix mcopy(nrows,ncols);
    
    size_type row = 1;
    iterator vec = rowvecs.begin();
    for (; vec!=rowvecs.end(); vec++, row++)
    	    mcopy.WriteRow(*vec,row);

    unsigned int error = m.IsNotEqualTo(mcopy);

    TestResults("ReadRow and WriteRow ?",error);
}
//.............................................................................
    
void CheckSVD(const BTL_Matrix& m)
{
    cout<<"\n###################\n"
    	 << "### Check SVD  ####\n"
    	 << "###            ####\n"
    	 << "###################\n\n";

    typedef BTL_Matrix::size_type size_type;

    // Make copy of original
    //
    BTL_Matrix mcopy = m;	 
    	 
    // Create BTL_Vector and Matrix to store the results
    //
    size_type ncols = m.ReadNCols();
    BTL_Vector L(ncols);
    BTL_Matrix V(ncols,ncols);
    
    // Calculuate SVD
    //	 
    mcopy.SVD(L,V);
    	
    cout << "Dimensions of input Matrix: " << m.ReadNRows() 
         << " x " << m.ReadNCols() << "\n\n";
    
    // Test svd product equals original matrix
    //
    // Make BTL_Vector L into diagonal Matrix

    BTL_Matrix matL(ncols,ncols);
    // extra curly brackets are for the benefit of SGI CC version 7.1
    {for (size_type i=1; i<=ncols; i++)
    	matL(i,i) = L(i);}
    
    	// Calculate SVD product    
    BTL_Matrix svd = mcopy * matL % V;
    
    	// Calculate and report error
    unsigned int error = m.IsNotEqualTo(svd);
    TestResults("A = U * matL % V ?",error);
}  
//.............................................................................

int main()
{

    // Small test square matrix


    BTL_Matrix matB(-1.0,  7.0, 9.2,
    	         8.7, -0.8, 9.0,
    	         4.0, -9.8, 0.4);
    	         
    BTL_Matrix matD(1,1);
    matD(1,1) = 8.7;

    BTL_Matrix matPosDef(4,2,14,
                     2,17,-5,
                     14,-5,83);

#include "../testCode/p4ins.dat"

    BTL_Matrix matA(p4ins_A,163,3);
    BTL_Matrix matC(p4ins_C,163,3);
    BTL_Vector point(-4.0,0.9,100), axis(78,-3,90);

    BTL_Matrix matE(10,3);
    for (int row=1; row<=10; row++)
    	for (int col=1; col<=3; col++)
    	    matE(row,col) = matA(row,col);
//    BTL_Matrix matF = matA.Transpose();

    CheckConstrct1();
    CheckConstrct2(matA);
    CheckRotateMat(point,axis,21.5);
    CheckSquareInverse(matA);
    CheckUnity(matA);
    CheckOperators1(matA);
    CheckOperators2(matD);
    CheckTranspose(matA);
    CheckProperty1(matA,matC);
    CheckProperty2(matA,matB,matC);
    CheckProperty3(matA,matB,matC);
    CheckProperty4(matA,matB,matC);
    CheckProperty5(matA,matC);
    CheckTransMult(matA);
    CheckEigens(matA);
    CheckCholesky(matPosDef);
    CheckRow(matA);
    CheckSVD(matA);
    CheckSVD(matB);
    CheckInverse(matA);

    return 0;
}
