
#if !defined(MOLECULE_H)
#define MOLECULE_H

#include "btl/XYZProcessor.h"
#include "btl/GraphWithEdges.h"

#include "vtkRenderMaster.h"
#include "vtkRenderWindow.h"
#include "vtkRenderer.h"
#include "vtkRenderWindowInteractor.h"
#include "vtkSphereSource.h"
#include "vtkPolyMapper.h"
#include "vtkCylinderSource.h"
#include "vtkActor.h"
#include "vtkVRMLExporter.h"
#include <math.h>
#include <vector> // include STL vector (to be covered in the last lecture)


//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// Abstract base class for displayable objects.
//

class Displayable
{
private:

    unsigned int id;
    static unsigned int nextId;

protected:

    vtkActor *actor;

public:

    // Make DisplayWindow class a friend so that it can access the actor object
    // and add it to the scene
    //    
    friend class DisplayWindow;
    
    // Default constructor
    //
    Displayable() { actor = new vtkActor; id = nextId++; }
    
    // Copy constructor
    //
    Displayable(const Displayable& d) 
    { 
    	actor = new vtkActor(*(d.actor)); 
	id = d.id;
    }

    // Default destructor (MUST be virtual)
    //
    virtual
    ~Displayable() {}
//    { delete actor; }
    
    // Get id number
    unsigned int
    GetId() const { return id; }
    
    // Abstract function to toggle between low res and high res.
    //
    virtual void 
    ToggleResolution() = 0;
    
    // Output operator (print function) (N.B. friend, so not inherited)
    //
    friend ostream& operator<<(ostream& os, const Displayable& d);
};


//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// This class can be used to create a new graphics window with one rendered
// scene within it.
//

class DisplayWindow
{
public:

    typedef vector<Displayable*> Container; 	// Extendable array of pointers

private:

    vtkRenderMaster rm;     	    	    	// The graphics coordinator
    vtkRenderWindow 	      	*renWin;    	// Window object
    vtkRenderer		      	*ren; 	    	// Renderer object
    vtkRenderWindowInteractor 	*iren;	    	// Mouse and keyboard interactor
    Container     	    	displayables;	// Vector of pointers to the
    	    	    	    	    	    	// displayable objects in the
    	    	    	    	    	    	// scene

public:

    // Create the window
    //
    DisplayWindow(); 
    
    // Export view to a VRML file
    //     
    void
    Export(char *fileName);
    
    // Add an 'actor' (non-speaking part) to the scene
    //
    void
    AddDisplayable(Displayable *displayable) 
    {	
    	displayables.push_back(displayable); 
    	ren->AddActors(displayable->actor);
    }
    
    // Render scene
    //
    void
    Render() { renWin->Render(); }
    
    // Render scene then start the interactor
    //
    void
    Start();
    
    // Toggle the resolution of each of the displayable objects displayed 
    // within this window.
    //
    // This function is a friend to this class and is therefore not part of
    // this class but can access its private parts. (This function has 
    // been made a friend for vtk specific reasons.)
    //
    friend void
    ToggleResolution(void *d); 

    // Output operator (print function)
    //
    friend ostream& operator<<(ostream& os, const DisplayWindow& w);
};

//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA


// Simple atom class - models atom as a sphere actor with a fixed radius and
// a variable position.

class Atom : public Displayable
{

    friend class Molecule;

private:

    vtkSphereSource 	*sphere;    	    	// Describes geometry of sphere 
    vtkPolyMapper   	*map;	    	    	// Maps actor onto hardware
    
    int     	    	id;
    char    	    	element[2];
    float     	    	position[3];  	    	// x,y,z coordinates of atom

    static int      	highRes, lowRes;    	// resolution levels for all
    	    	    	    	    	    	// Atoms
    bool    	    	bestResolution;     	// Controls resolution of this
    	    	    	    	    	        // Atom
    void						
    GetElementColours(char a1, char a2, float& r, float& g, float& b);
    
    // Define the initial size and resolution of the Atom sphere
    // 
    void
    Initialise();

public:

    // Create white atom at the origin
    //
    Atom();

    // Create an atom at a given position with a given colour
    //
    Atom(const float x, const float y, const float z, 
         const float r, const float g, const float b);
	 
    // Copy constructor
    //
    Atom(const Atom& a);
    
    // Destructor
    //
//     virtual
//     ~Atom() { DEBUG(GetId()); delete sphere; delete map; }
    
    // Assignment operator
    //
    void
    operator=(const Atom& a);

    // Returns x coordinate
    //
    float
    GetX() const { return position[0]; }
    
    // Returns y coordinate
    //
    float
    GetY() const { return position[1]; }
    
    // Returns z coordinate
    //
    float
    GetZ() const { return position[2]; }
    
    // Change position of atom
    //
    void
    SetPosition(const float x, const float y, const float z)
    { 
    	position[0] = x; 
	position[1] = y; 
	position[2] = z;
    	actor->SetPosition(x,y,z);
    }
    
    // Change position of atom
    //
    void
    SetElement(char a1, char a2) 
    { 
    	element[0] = a1; 
	element[1] = a2;
	float r,g,b;
	GetElementColours(a1,a2,r,g,b);
	actor->GetProperty()->SetColor(r,g,b);
    }

    // Toggle resolution - smooth motion or best resolution
    //
    void
    ToggleResolution();

    // Output operator (print function)
    //
    friend ostream& operator<<(ostream& os, const Atom& atom);
};

//VVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVVV
//||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
//AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA

// Simple bond class - models bond as green cylinder of fixed radius. Need two
// Atom objects to create a bond.

class Bond : public Displayable
{
private:

    vtkCylinderSource 	*cylinder;  	    // Cylinder geometry object
    vtkPolyMapper   	*map;	    	    // Maps actor onto hardware
    
    static int      	highRes, lowRes;    // resolution levels for all
            	    	    	    	    // Atoms
    bool    	    	bestResolution;     // Controls resolution of this
					    // Atom

    // Define the size and resolution of the bond cylinder
    // 
    void
    Initialise();
        
public:

    Bond(){}

    // Create bond between two atoms
    //
    Bond(Atom *atom1, Atom *atom2, float bondLength);
    
//     // Copy constructor (shallow copies if vtk objects).
//     //
//     Bond(const Bond& b)
//     {
//     	cylinder=new vtkCylinderSource(*(b.cylinder));
// 	map = new vtkPolyMapper(*(b.map));
// 	bestResolution = b.bestResolution; 
//     }
    		
    
    // Destructor
    //
//     virtual
//     ~Bond() { DEBUG(GetId()); delete cylinder; delete map; }

    // Toggle between smooth motion and high resolution
    //
    void
    ToggleResolution();
    
    // Output operator (print function)
    //
    friend ostream& operator<<(ostream& os, const Bond& bond);
};

class Molecule 
{
public:

    typedef vector<Atom*> C1;
    typedef vector<Bond*> C2;
    
    typedef BTL_GraphWithEdges< BTL_InD<C1>, BTL_InD<C2> > AtomsAndBonds;
//    typedef BTL_Graph6< vector<Atom>, vector<Bond> > AtomsAndBonds;

private:

    AtomsAndBonds atomsAndBonds;
    
    void 
    CreateBonds();
    
public:
  
    Molecule(BTL_XYZProcessor& p);
    
    ~Molecule();
       
    void
    AddToWindow(DisplayWindow& win);

    friend ostream& 
    operator<<(ostream& os, const Molecule& m);
};    


#endif
